"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const workspace_tools_1 = require("workspace-tools");
const taskId_1 = require("./taskId");
const generateTask_1 = require("./generateTask");
const path_1 = __importDefault(require("path"));
const cacheTasks_1 = require("../cache/cacheTasks");
const cosmiconfig_1 = require("cosmiconfig");
const ConfigModuleName = "lage";
function filterPackages(context) {
    const { allPackages, scope, deps: withDeps } = context;
    let scopes = [].concat(scope);
    let scopedPackages = scopes && scopes.length > 0
        ? workspace_tools_1.getScopedPackages(scopes, allPackages)
        : Object.keys(allPackages);
    if (withDeps) {
        scopedPackages = scopedPackages.concat(workspace_tools_1.getTransitiveDependencies(scopedPackages, allPackages));
    }
    return scopedPackages;
}
function getPipeline(pkg, context) {
    const { allPackages, defaultPipeline } = context;
    const info = allPackages[pkg];
    const results = cosmiconfig_1.cosmiconfigSync(ConfigModuleName).search(path_1.default.dirname(info.packageJsonPath));
    let pipeline = defaultPipeline;
    if (results && results.config) {
        pipeline = results.config.pipeline;
    }
    return pipeline;
}
/**
 * Gather all the task dependencies defined by the "pipeline" setting, generates a list of edges
 * @param targetTask
 * @param pipeline
 */
function generateTaskDepsGraph(targetTask, pipeline) {
    const queue = [targetTask];
    const visited = new Set();
    const graph = [];
    while (queue.length > 0) {
        const task = queue.shift();
        if (!visited.has(task)) {
            visited.add(task);
            if (Array.isArray(pipeline[task])) {
                if (pipeline[task].length > 0) {
                    for (const depTask of pipeline[task]) {
                        graph.push([depTask, task]);
                        queue.push(depTask);
                    }
                }
                else {
                    graph.push(["", task]);
                }
            }
        }
    }
    return graph;
}
/**
 * identify and create a realized task dependency map (discovering)
 *
 * This function will traverse the package dependency graph, and will end up traverse the task depenendencies within the same package (2 layered traversal)
 */
function discoverTaskDeps(context) {
    const { allPackages, command } = context;
    const filteredPackages = filterPackages(context);
    console.log(`Packages in scope: ${filteredPackages.join(",")}`);
    // initialize a queue for a breadth first approach
    const traversalQueue = filteredPackages;
    const visited = new Set();
    const dependentMap = workspace_tools_1.getDependentMap(allPackages);
    while (traversalQueue.length > 0) {
        const pkg = traversalQueue.shift();
        if (!visited.has(pkg)) {
            visited.add(pkg);
            // get pipeline
            const pipeline = getPipeline(pkg, context);
            // establish task graph; push dependents in the traversal queue
            const depTaskGraph = generateTaskDepsGraph(command, pipeline);
            for (const [from, to] of depTaskGraph) {
                const dependentPkgs = dependentMap.get(pkg);
                const toTaskId = taskId_1.getTaskId(pkg, to);
                if (from.startsWith("^") && dependentPkgs !== undefined) {
                    // add task dep from all the package deps within repo
                    for (const depPkg of dependentPkgs) {
                        const fromTaskId = taskId_1.getTaskId(depPkg, from.slice(1));
                        createDep(fromTaskId, toTaskId, context);
                    }
                    // now push the dependents in the traversal queue
                    traversalQueue.push(pkg);
                }
                else {
                    const fromTaskId = taskId_1.getTaskId(pkg, from);
                    // add task dep from same package
                    createDep(fromTaskId, toTaskId, context);
                }
            }
        }
    }
}
exports.discoverTaskDeps = discoverTaskDeps;
function isValidTaskId(taskId, allPackages) {
    const [pkg, task] = taskId_1.getPackageTaskFromId(taskId);
    return (taskId === "" ||
        [cacheTasks_1.ComputeHashTask, cacheTasks_1.CachePutTask, cacheTasks_1.CacheFetchTask].includes(task) ||
        Object.keys(allPackages[pkg].scripts || {}).includes(task));
}
function createDep(fromTaskId, toTaskId, context) {
    const { tasks, taskDepsGraph, allPackages } = context;
    if (!isValidTaskId(fromTaskId, allPackages)) {
        return;
    }
    taskDepsGraph.push([fromTaskId, toTaskId]);
    if (!tasks.get(fromTaskId)) {
        tasks.set(fromTaskId, () => generateTask_1.generateTask(fromTaskId, context));
    }
    if (!tasks.get(toTaskId)) {
        tasks.set(toTaskId, () => generateTask_1.generateTask(toTaskId, context));
    }
}
